from __future__ import annotations
from logging import getLogger
from pathlib import Path
import errno
import os

# external
import xarray as xr

# package
from assetra.units import (
    EnergyUnit,
    RESPONSIVE_UNIT_TYPES,
    NONRESPONSIVE_UNIT_TYPES,
)

LOG = getLogger(__name__)


class EnergySystem:
    """Class responsible for managing unit datasets (built energy systems)

    Args:
        unit_datasets (dict[Type : xr.Dataset]) : Mapping from derived energy
            unit type to its associated unit dataset
    """

    def __init__(self, unit_datasets: dict[type : xr.DataArray] = {}):
        self._unit_datasets = unit_datasets

        # check input
        for unit_type in unit_datasets:
            if unit_type not in (
                RESPONSIVE_UNIT_TYPES + NONRESPONSIVE_UNIT_TYPES
            ):
                LOG.error(
                    "Constructing energy system with invalid unit dataset"
                )
                raise RuntimeWarning

    @property
    def size(self) -> float:
        """
        Total number of units in system
        """
        return sum(d.sizes["energy_unit"] for d in self._unit_datasets.values())

    @property
    def system_capacity(self) -> float:
        """
        Total nameplate capacity of a system
        """
        return sum(
            float(d["nameplate_capacity"].sum())
            for d in self._unit_datasets.values()
        )

    @property
    def unit_datasets(self) -> dict[type : xr.Dataset]:
        """
        Mapping from derived energy unit type to unit dataset
        """
        return self._unit_datasets

    def get_system_by_type(self, unit_type: list[type] | type) -> EnergySystem:
        """Return a system comprised of the subset of unit datasets
        corresponding to one (or more) energy unit types.

        For example, get a [sub]system with only the responsive or non-responsive
        units of a system

        Args:
            unit_type (list[type] | type): Either a derived energy unit type or
                a list of derived energy unit types.

        Returns:
            EnergySystem: An energy system whose unit datasets are a sub-set of
                of this energy system.
        """
        if isinstance(unit_type, (list, tuple)):
            unit_datasets = {
                ut: ud
                for ut, ud in self._unit_datasets.items()
                if ut in unit_type
            }
            return EnergySystem(unit_datasets)
        elif unit_type in RESPONSIVE_UNIT_TYPES + NONRESPONSIVE_UNIT_TYPES:
            return EnergySystem({unit_type: self._unit_datasets[unit_type]})

    def save(self, directory: Path, overwrite=False) -> None:
        """Save energy system to a directory. Unit datasets are saved as netcdf
        files

        Args:
            directory (Path): Path to which energy system is saved. This path
                should either be empty or not exist yet.
            overwrite (bool, optional): _description_. Defaults to False.
        """
        directory.mkdir(parents=True, exist_ok=overwrite)
        for unit_type, dataset in self._unit_datasets.items():
            dataset_file = Path(directory, unit_type.__name__ + ".assetra.nc")
            dataset.to_netcdf(dataset_file)

    def load(self, directory: Path) -> None:
        """Load energy system from a saved directory

        Args:
            directory (Path): Path from which energy system is loaded. This
                should be the same path passed to EnergySystem.save
        """
        if not directory.exists():
            raise FileNotFoundError()

        self._unit_datasets = dict()

        for unit_type in NONRESPONSIVE_UNIT_TYPES + RESPONSIVE_UNIT_TYPES:
            dataset_file = Path(directory, unit_type.__name__ + ".assetra.nc")

            if dataset_file.exists():
                LOG.info("Found unit dataset: " + str(dataset_file.resolve()))
                self._unit_datasets[unit_type] = xr.open_dataset(dataset_file)


class EnergySystemBuilder:
    """Class responsible for managing energy units and building energy systems"""

    def __init__(self):
        self._energy_units = []

    @property
    def energy_units(self) -> tuple[EnergyUnit]:
        """
        Energy units added to builder object
        """
        return tuple(self._energy_units)

    @property
    def size(self) -> int:
        """
        Number of energy units added to builder object
        """
        return len(self._energy_units)

    def add_unit(self, energy_unit: EnergyUnit) -> None:
        """Add an energy unit to the system builder object.

        Args:
            energy_unit (EnergyUnit): Energy unit to add to system builder

        Raises:
            RuntimeWarning: Invalid energy unit type added to system builder
            RuntimeWarning: Duplicate unit added to energy system builder
        """
        # check for valid energy unit
        if (
            type(energy_unit)
            not in NONRESPONSIVE_UNIT_TYPES + RESPONSIVE_UNIT_TYPES
        ):
            LOG.warning("Invalid type added to energy system builder")
            raise RuntimeWarning()

        # check for duplicates
        if energy_unit.id in [u.id for u in self._energy_units]:
            LOG.warning("Duplicate unit ID added to energy system builder")
            raise RuntimeWarning()

        # add unit to internal list
        self._energy_units.append(energy_unit)

    def remove_unit(self, energy_unit: EnergyUnit) -> None:
        """Remove an energy unit from the system builder object.

        Args:
            energy_unit (EnergyUnit): Energy unit to remove from system builder
        """
        try:
            self._energy_units.remove(energy_unit)
        except KeyError:
            LOG.warning("Unit to remove not found in energy system builder")

    def build(self) -> EnergySystem:
        """Return a populated EnergySystem instance. Take energy units added to
        the builder object, compile each unit type into a unit dataset (fleet),
        and instantiate an EnergySystem with the resulting unit dataset
        dictionary.

        This is the recommended method to instantiate EnergySystem objects

        Returns:
            EnergySystem: Populated energy system instance
        """
        unit_datasets = dict()

        # populate unit datasets
        for unit_type in NONRESPONSIVE_UNIT_TYPES + RESPONSIVE_UNIT_TYPES:
            # get unit by type
            units = [
                unit for unit in self.energy_units if type(unit) is unit_type
            ]

            # get unit dataset
            if len(units) > 0:
                unit_datasets[unit_type] = unit_type.to_unit_dataset(units)

        return EnergySystem(unit_datasets)

    @staticmethod
    def from_energy_system(energy_system: EnergySystem) -> EnergySystemBuilder:
        """Return a populated EnergySystemBuilder instance. Take energy unit
        datasets from an energy system, convert datasets into
        individual energy units, and add units to a new EnergySystemBuilder
        object

        This is the inverse to EnergySystem.build, and is useful for
        modifying energy systems which have been loaded directly from file
        using the EnergySystem.load function

        Args:
            energy_system (EnergySystem): Populated energy system

        Returns:
            EnergySystemBuilder: Populated builder instance
        """
        builder = EnergySystemBuilder()

        for unit_type, unit_dataset in energy_system.unit_datasets.items():
            units = unit_type.from_unit_dataset(unit_dataset)
            for unit in units:
                builder.add_unit(unit)

        return builder



from __future__ import annotations
from abc import abstractmethod, ABC
from logging import getLogger
from dataclasses import dataclass

# external
import numpy as np
import xarray as xr

LOG = getLogger(__name__)

# ENERGY UNIT(S)
MAX_CHUNK_SIZE = int(1e8)


@dataclass(frozen=True)
class EnergyUnit(ABC):
    """Abstract base class for all energy units.

    Energy units are the fundamental building blocks of energy systems in
    the assetra model. This base class defines an interface which allows the
    assetra model to save/load pre-existing energy systems from files and run
    probabilistic simulations with unique energy unit types.

    Args:
        id (int): Unique identifying number, used to ensure energy units are
            not added more than once to an energy system
        nameplate_capacity (float) : Nameplate capacity of the energy unit in
            units of power (to be kept consistend between units). For some
            units defining the nameplate capacity may not make physical sense,
            e.g. demand units, in which case the nameplate capacity should be
            set to zero.
    """
    id: int
    nameplate_capacity: float

    @staticmethod
    @abstractmethod
    def to_unit_dataset(units: list[EnergyUnit]) -> xr.Dataset:
        """Convert a list of energy units of the derived class type into an
        xarray dataset.

        For different energy units, different dataset
        dimensions and coordinates may be appropriate.

        Args:
            units (list[EnergyUnit]): List of of energy units of the derived
                class type

        Returns:
            xr.Dataset: Dataset storing sufficient information to (1) fully
                reconstruct the list of energy units from which it is created
                and (2) generate hourly capacity time series with the
                EnergyUnit.get_probabilistic_capacity_matrix function
        """
        pass

    @staticmethod
    @abstractmethod
    def from_unit_dataset(unit_dataset: xr.Dataset) -> list[EnergyUnit]:
        """Convert a unit dataset to a list of energy units of the derived
        energy unit type.

        This is the inverse to the derived EnergyUnit.to_unit_dataset function

        Args:
            unit_dataset (xr.Dataset): Unit dataset with structure and content
                defined in the derived EnergyUnit.to_unit_dataset function

        Returns:
            list[EnergyUnit]: List of energy units of the derived class type
        """

    @staticmethod
    @abstractmethod
    def get_probabilistic_capacity_matrix(
        unit_dataset: xr.Dataset, net_hourly_capacity_matrix: xr.DataArray
    ) -> xr.DataArray:
        """Return probabilistic hourly capacity matrix for a fleet of energy
        units of the derived energy unit type.

        Take the unit dataset and create a matrix representing the total hourly
        capacity of all energy units for some number of monte carlo trials. The
        hours and number of trials should match the net hourly capacity matrix.

        Args:
            unit_dataset (xr.Dataset): Unit dataset for the derived energy unit
                type, e.g. generated with the derived
                EnergyUnit.to_unit_dataset function
            net_hourly_capacity_matrix (xr.DataArray): Probabilistic net hourly
                capacity matrix with dimensions (trials, time) and shape
                (# of trials, # of hours)

        Returns:
            xr.DataArray: Combined hourly capacity for all units in the unit
                dataset for a determined number of Monte Carlo trials. The
                dimensions and coordinates of this matrix should match the net
                hourly capacity matrix
        """
        pass


@dataclass(frozen=True)
class StaticUnit(EnergyUnit):
    """Derived energy unit class.

    A static energy unit is neither stochastic nor responsive. A single
    hourly capacity profile is used in all probabilistic capacity trials.
    For example, a historical demand profile be fully accounted for in all
    trials of a probabilistic simulation.

    Args:
        id (int): Unique identifying number
        nameplate_capacity (float) : Nameplate capacity of the energy unit in
            units of power
        hourly_capacity (xr.DataArray) : Hourly capacity contained in DataArray
            with dimension (time) and datetime coordinates.
    """

    hourly_capacity: xr.DataArray

    @staticmethod
    def to_unit_dataset(units: list[StaticUnit]) -> xr.Dataset:
        """Convert a list of static energy units into an xarray dataset

        Args:
            units (list[StaticUnit]): List of of static energy units

        Returns:
            xr.Dataset: Dataset with dimensions (energy_unit, time) and
                variables (nameplate_capacity[energy_unit],
                hourly_capacity[energy_unit, time]). Coordinates for the
                energy_unit and time dimensions are energy unit IDs and
                hourly datetime indices, respectively.
        """
        # build dataset
        unit_dataset = xr.Dataset(
            data_vars=dict(
                nameplate_capacity=(
                    ["energy_unit"],
                    [unit.nameplate_capacity for unit in units],
                ),
                hourly_capacity=(
                    ["energy_unit", "time"],
                    [unit.hourly_capacity for unit in units],
                ),
            ),
            coords=dict(
                energy_unit=[unit.id for unit in units],
                time=units[0].hourly_capacity.time if len(units) > 0 else [],
            ),
        )

        return unit_dataset

    @staticmethod
    def from_unit_dataset(unit_dataset: xr.Dataset) -> list[StaticUnit]:
        """Convert a static unit dataset to a list of static energy units.

        This is the inverse StaticUnit.to_unit_dataset function

        Args:
            unit_dataset (xr.Dataset): Unit dataset with structure and content
                defined in the derived StaticUnit.to_unit_dataset function

        Returns:
            list[StaticUnit]: List of static energy units
        """
        # build list
        units = []

        for id in unit_dataset.energy_unit:
            units.append(
                StaticUnit(
                    int(id),
                    int(unit_dataset.nameplate_capacity.loc[id]),
                    unit_dataset.hourly_capacity.loc[id],
                )
            )

        return units

    @staticmethod
    def get_probabilistic_capacity_matrix(
        unit_dataset: xr.Dataset, net_hourly_capacity_matrix: xr.DataArray
    ) -> xr.DataArray:
        """Return probabilistic hourly capacity matrix for a static unit
        dataset.

        For static units, combine hourly capacity profiles for all energy units
        in the unit dataset and broadcast the result across all trials

        Args:
            unit_dataset (xr.Dataset): Static unit dataset, as generated by
                StaticUnit.to_unit_dataset function
            net_hourly_capacity_matrix (xr.DataArray): Probabilistic net hourly
                capacity matrix with dimensions (trials, time) and shape
                (# of trials, # of hours)

        Returns:
            xr.DataArray: Combined hourly capacity for all units in the unit
                dataset with the same dimensions and shape as net hourly
                capacity matrix
        """
        # time-indexing
        unit_dataset = unit_dataset.sel(time=net_hourly_capacity_matrix.time)

        # sum across capacity units
        probabilistic_capacity_matrix = unit_dataset["hourly_capacity"].sum(
            dim="energy_unit"
        )

        # to xarray
        probabilistic_capacity_matrix = (
            xr.zeros_like(net_hourly_capacity_matrix)
            + probabilistic_capacity_matrix
        )

        return probabilistic_capacity_matrix


@dataclass(frozen=True)
class StochasticUnit(EnergyUnit):
    """Derived energy unit class.

    A stochastic energy unit uses time-varying forced outage rates to sample
    indepenedent outages throughout the simulation period. Stochastic units are
    non-responsive, meaning that while hourly capacity profiles vary between
    trials in a probabilistic simulation, the profiles do not depend on system
    conditions and only need to be sampled once

    Args:
        id (int): Unique identifying number
        nameplate_capacity (float) : Nameplate capacity of the energy unit in
            units of power
        hourly_capacity (xr.DataArray) : Hourly capacity contained in DataArray
            with dimension (time) and datetime coordinates
        hourly_forced_outage_rate (xr.DataArray) : Hourly forced outage rate
            as decimal percents (e.g. 5% -> 0.05) contained in DataArray with
            dimension (time) and datetime coordinates. Should be a parallel
            matrix to hourly_capacity
    """

    hourly_capacity: xr.DataArray
    hourly_forced_outage_rate: xr.DataArray

    @staticmethod
    def to_unit_dataset(units: list[StochasticUnit]):
        """Convert a list of stochastic energy units into an xarray dataset

        Args:
            units (list[StochasticUnit]): List of of stochastic energy units

        Returns:
            xr.Dataset: Dataset with dimensions (energy_unit, time) and
                variables (nameplate_capacity[energy_unit],
                hourly_capacity[energy_unit, time]),
                hourly_forced_outage_rate[energy_unit, time]. Coordinates for
                the energy_unit and time dimensions are energy unit IDs and
                hourly datetime indices, respectively.
        """
        unit_dataset = xr.Dataset(
            data_vars=dict(
                nameplate_capacity=(
                    ["energy_unit"],
                    [unit.nameplate_capacity for unit in units],
                ),
                hourly_capacity=(
                    ["energy_unit", "time"],
                    [unit.hourly_capacity for unit in units],
                ),
                hourly_forced_outage_rate=(
                    ["energy_unit", "time"],
                    [unit.hourly_forced_outage_rate for unit in units],
                ),
            ),
            coords=dict(
                energy_unit=[unit.id for unit in units],
                time=units[0].hourly_capacity.time if len(units) > 0 else [],
            ),
        )
        return unit_dataset

    @staticmethod
    def from_unit_dataset(unit_dataset: xr.Dataset) -> list[StochasticUnit]:
        """Convert a stochastic unit dataset to a list of stochastic energy units.

        This is the inverse to StochasticUnit.to_unit_dataset function

        Args:
            unit_dataset (xr.Dataset): Unit dataset with structure and content
                defined in the derived StochasticUnit.to_unit_dataset function

        Returns:
            list[StochasticUnit]: List of stochastic units
        """
        # build list
        units = []

        for id in unit_dataset.energy_unit:
            units.append(
                StochasticUnit(
                    id,
                    unit_dataset.nameplate_capacity.loc[id],
                    unit_dataset.hourly_capacity.loc[id],
                    unit_dataset.hourly_forced_outage_rate.loc[id],
                )
            )

        return units

    @staticmethod
    def get_probabilistic_capacity_matrix(
        unit_dataset: xr.Dataset, net_hourly_capacity_matrix: xr.DataArray
    ) -> xr.DataArray:
        """Return probabilistic hourly capacity matrix for a stochastic unit
        dataset.

        For stochastic units, sample hourly independent outages in for units
        in all trials. Outages are sampled hourly for every unit and trial.
        Random numbers are drawn from the range 0 to 1, and where samples are
        less than the hourly forced outage rate, the effective capacity of
        that energy unit in that hour and trial is set to 0. The probabilistic
        capacity matrix is the aggregation of sampled capacities across energy
        units

        Args:
            unit_dataset (xr.Dataset): Static unit dataset, as generated by
                StaticUnit.to_unit_dataset function
            net_hourly_capacity_matrix (xr.DataArray): Probabilistic net hourly
                capacity matrix with dimensions (trials, time) and shape
                (# of trials, # of hours)

        Returns:
            xr.DataArray: Combined hourly capacity for all units in the unit
                dataset with the same dimensions and shape as net hourly
                capacity matrix
        """
        # time-indexing
        unit_dataset = unit_dataset.sel(time=net_hourly_capacity_matrix.time)

        chunk_size = 1 + MAX_CHUNK_SIZE // (
            net_hourly_capacity_matrix.sizes["trial"]
            * net_hourly_capacity_matrix.sizes["time"]
        )
        LOG.info("Using chunk size " + str(chunk_size))

        # Initialize the probabilistic capacity matrix
        probabilistic_capacity_matrix = xr.zeros_like(
            net_hourly_capacity_matrix
        )

        # Loop over the energy unit dimension in chunks
        for unit_idx in range(0, unit_dataset.sizes["energy_unit"], chunk_size):
            unit_idx_end = min(
                unit_idx + chunk_size, unit_dataset.sizes["energy_unit"]
            )

            LOG.info(
                "Sampling outages for units "
                + str(unit_idx + 1)
                + "-"
                + str(unit_idx_end)
                + " of "
                + str(unit_dataset.sizes["energy_unit"])
            )

            chunk = unit_dataset.isel(energy_unit=slice(unit_idx, unit_idx_end))

            chunk_prob_matrix = np.where(
                np.random.random_sample(
                    (
                        net_hourly_capacity_matrix.sizes["trial"],
                        chunk.sizes["energy_unit"],
                        net_hourly_capacity_matrix.sizes["time"],
                    )
                )
                > chunk["hourly_forced_outage_rate"].values,
                chunk["hourly_capacity"].values,
                0,
            ).sum(axis=1)

            # Update the main probabilistic capacity matrix with the results from the chunk
            probabilistic_capacity_matrix += chunk_prob_matrix

        return probabilistic_capacity_matrix


@dataclass(frozen=True)
class StorageUnit(EnergyUnit):
    """Derived energy unit class.

    A storage unit is a state-dependent, responsive energy unit. The available
    capacity of a storage unit depends on its state of charge and on the needs
    of the system. As opposed to static and stochastic units, which require
    hourly time series, storage unit operation is characterized by a handful of
    scalar parameters

    Args:
        id (int): Unique identifying number
        nameplate_capacity (float) : Nameplate capacity in units of power. For
            storage, typically the discharge rate
        charge_rate (float) : Charge rate in units of power
        discharge_rate (float) : Discharge rate in units of power
        charge_capacity (float) : Maximum charge capacity in units of energy
        roundtrip_efficiency (float) : Roundtrip efficiency as decimal percent
    """

    charge_rate: float
    discharge_rate: float
    charge_capacity: float
    roundtrip_efficiency: float

    def _get_hourly_capacity(
        charge_rate: float,
        discharge_rate: float,
        charge_capacity: float,
        roundtrip_efficiency: float,
        net_hourly_capacity: xr.DataArray,
    ) -> xr.DataArray:
        """Greedy storage dispatch

        Args:
            charge_rate (float) : Charge rate in units of power
            discharge_rate (float) : Discharge rate in units of power
            charge_capacity (float) : Maximum charge capacity in units of
                energy
            roundtrip_efficiency (float) : Roundtrip efficiency as decimal
                percent
            net_hourly_capacity (xr.DataArray): Net capacity contained in
                DataArray with dimension (time) and hourly datetime
                coordinates

        Returns:
            xr.DataArray: Hourly capacity contained in DataArray with same
                shape as net hourly capacity
        """
        # TODO skip irrelevant days for average-case speed-up?
        # initialize full storage unit
        efficiency = roundtrip_efficiency**0.5

        def charge_storage(excess_capacity: float, current_charge: float):
            capacity = -min(
                charge_rate,
                (charge_capacity - current_charge) / efficiency,
                excess_capacity,
            )
            current_charge -= capacity * efficiency

            return capacity, current_charge

        def discharge_storage(unmet_demand: float, current_charge: float):
            capacity = min(
                discharge_rate / efficiency,
                current_charge,
                unmet_demand / efficiency,
            )
            current_charge -= capacity

            return capacity * efficiency, current_charge

        def dispatch_storage(net_hourly_capacity: float):
            # initialize storage unit as full
            current_charge = float(charge_capacity)

            for net_capacity in net_hourly_capacity.values:
                capacity = 0
                if net_capacity < 0:
                    if current_charge > 0:
                        # unmet demand and avaiable charge
                        capacity, current_charge = discharge_storage(
                            -net_capacity, current_charge
                        )
                elif current_charge < charge_capacity:
                    # excess capacity and not fully charged
                    capacity, current_charge = charge_storage(
                        net_capacity, current_charge
                    )
                yield capacity

        # simulate dispatch
        hourly_capacity = net_hourly_capacity.copy(
            data=[
                capacity for capacity in dispatch_storage(net_hourly_capacity)
            ]
        )

        return hourly_capacity

    @staticmethod
    def to_unit_dataset(units: list[StorageUnit]) -> xr.Dataset:
        """Convert a list of storage units into an xarray dataset

        Args:
            units (list[StorageUnit]): List of of storage energy units

        Returns:
            xr.Dataset: Dataset with dimensions (energy_unit) and
                variables (nameplate_capacity[energy_unit],
                charge_rate[energy_unit], discharge_rate[energy_unit],
                charge_capacity[energy_unit], roundtrip_efficiency[energy_unit]
                hourly_forced_outage_rate[energy_unit, time]). Coordinates for
                the energy_unit dimension are energy unit IDs
        """
        # build dataset
        unit_dataset = xr.Dataset(
            data_vars=dict(
                nameplate_capacity=(
                    ["energy_unit"],
                    [unit.nameplate_capacity for unit in units],
                ),
                charge_rate=(
                    ["energy_unit"],
                    [unit.charge_rate for unit in units],
                ),
                discharge_rate=(
                    ["energy_unit"],
                    [unit.discharge_rate for unit in units],
                ),
                charge_capacity=(
                    ["energy_unit"],
                    [unit.charge_capacity for unit in units],
                ),
                roundtrip_efficiency=(
                    ["energy_unit"],
                    [unit.roundtrip_efficiency for unit in units],
                ),
            ),
            coords=dict(energy_unit=[unit.id for unit in units]),
        )

        return unit_dataset

    @staticmethod
    def from_unit_dataset(unit_dataset: xr.Dataset) -> list[StorageUnit]:
        """Convert a storage unit dataset to a list of storage units.

        This is the inverse StorageUnit.to_unit_dataset function

        Args:
            unit_dataset (xr.Dataset): Unit dataset with structure and content
                defined in the derived StorageUnit.to_unit_dataset function

        Returns:
            list[StorageUnit]: List of storage units
        """
        # build list
        units = []

        for id, nc, cr, dr, cc, re in zip(
            unit_dataset.energy_unit,
            unit_dataset.nameplate_capacity,
            unit_dataset.charge_rate,
            unit_dataset.discharge_rate,
            unit_dataset.charge_capacity,
            unit_dataset.roundtrip_efficiency,
        ):
            units.append(
                StorageUnit(
                    id=int(id),
                    nameplate_capacity=float(nc),
                    charge_rate=float(cr),
                    discharge_rate=float(dr),
                    charge_capacity=float(cc),
                    roundtrip_efficiency=float(re),
                )
            )

        return units

    @staticmethod
    def get_probabilistic_capacity_matrix(
        unit_dataset: xr.Dataset, net_hourly_capacity_matrix: xr.DataArray
    ) -> xr.DataArray:
        """Return probabilistic hourly capacity matrix for a storage unit
        dataset.

        For storage units, it is necessary to dispatch units every hour and
        iteration sequentially. The dispatch policy implemented in
        StorageUnit._get_hourly_capacity is a greedy policy to minimize
        expected unserved energy. Units are dispatched according to the order
        they appear in the unit dataset

        Args:
            unit_dataset (xr.Dataset): Storage unit dataset, as generated by
                StorageUnit.to_unit_dataset function
            net_hourly_capacity_matrix (xr.DataArray): Probabilistic net hourly
                capacity matrix with dimensions (trials, time) and shape
                (# of trials, # of hours)

        Returns:
            xr.DataArray: Combined hourly capacity for all units in the unit
                dataset with the same dimensions and shape as net hourly
                capacity matrix
        """
        units = StorageUnit.from_unit_dataset(unit_dataset)

        net_adj_hourly_capacity_matrix = net_hourly_capacity_matrix.copy()
        for idx, unit in enumerate(units):
            # print update
            LOG.info(
                "Dispatching storage unit "
                + str(idx)
                + " of "
                + str(len(units))
                + " in all hours"
            )
            for trial in net_adj_hourly_capacity_matrix:
                trial += StorageUnit._get_hourly_capacity(
                    unit.charge_rate,
                    unit.discharge_rate,
                    unit.charge_capacity,
                    unit.roundtrip_efficiency,
                    trial,
                )

        return net_adj_hourly_capacity_matrix - net_hourly_capacity_matrix


class DemandUnit(StaticUnit):
    """Derived energy unit class, providing a more meaningful interface for
    demand. Internally, demand is treated as a static unit with negative
    capacity and a nameplate rating of zero.

    This interface converts positive hourly demand into negative hourly
    capacity

    Args:
        id (int): Unique identifying number
        hourly_demand (xr.DataArray) : Hourly demand contained in DataArray
            with dimension (time) and datetime coordinates.
    """

    def __init__(self, id: int, hourly_demand: xr.DataArray):
        StaticUnit.__init__(
            self, id, nameplate_capacity=0, hourly_capacity=-hourly_demand
        )


# for successive simulations (e.g. ELCC), need to differentiate between
# responsive and non-responsive units.
#
# these lists also serve to track all "valid" units that can be added
# to an energy system
NONRESPONSIVE_UNIT_TYPES = [DemandUnit, StaticUnit, StochasticUnit]
RESPONSIVE_UNIT_TYPES = [StorageUnit]



from abc import ABC, abstractmethod

from assetra.simulation import ProbabilisticSimulation


class ResourceAdequacyMetric(ABC):
    """Class responsible for quantifying resource adequacy of a
    ProbabilisticSimulation object

    Args:
        simulation (ProbabilisticSimulation) : Simulation object to
            characterize.
    """

    def __init__(self, simulation: ProbabilisticSimulation):
        self.simulation = simulation

    @abstractmethod
    def evaluate(self) -> float:
        """Return resource adequacy of initialized simulation.

        Returns:
            float: Evaluated resource adequacy metric.
        """
        pass


class ExpectedUnservedEnergy(ResourceAdequacyMetric):
    """Derived ResourceAdequacyMetric class responsible for calculating
    expected unserved energy:

    "[The] total expected amount of unserved energy ... in a given study
    horizon."

    https://gridops.epri.com/Adequacy/metrics#Magnitude_Metrics

    Args:
        simulation (ProbabilisticSimulation) : Simulation object to
            characterize.
    """

    def evaluate(self) -> float:
        """Return expected unserved energy

        Returns:
            float: EUE in units of energy
        """
        hourly_unserved_energy = (
            self.simulation.net_hourly_capacity_matrix.where(
                self.simulation.net_hourly_capacity_matrix < 0, 0
            )
        )
        return float(
            -hourly_unserved_energy.sum()
            / hourly_unserved_energy.sizes["trial"]
        )


class LossOfLoadHours(ResourceAdequacyMetric):
    """Derived ResourceAdequacyMetric class responsible for calculating loss of
    load hours:

    "[The] expected count of event-hours per study horizon."

    https://gridops.epri.com/Adequacy/metrics#Loss_of_Load_Hours_.28LOLH.29

    Args:
        simulation (ProbabilisticSimulation) : Simulation object to
            characterize.
    """

    def evaluate(self) -> float:
        """Return loss of load hours

        Returns:
            float: LOLH in units of hours
        """
        hourly_outages = self.simulation.net_hourly_capacity_matrix < 0
        return float(hourly_outages.sum() / hourly_outages.sizes["trial"])


class LossOfLoadDays(ResourceAdequacyMetric):
    """Derived ResourceAdequacyMetric class responsible for calculating loss of
    load days:

    "[The] expected count of event-days per study horizon."

    https://gridops.epri.com/Adequacy/metrics#Loss_of_Load_Days_.28LOLD.2C_LOLEd.2Fyr.29

    Args:
        simulation (ProbabilisticSimulation) : Simulation object to
            characterize.
    """

    def evaluate(self) -> float:
        """Return loss of load days

        Returns:
            float: LOLD in units of days
        """
        hourly_outages = self.simulation.net_hourly_capacity_matrix < 0
        daily_outages = hourly_outages.resample(time="1D").max()
        return float(daily_outages.sum() / hourly_outages.sizes["trial"])


class LossOfLoadFrequency(ResourceAdequacyMetric):
    """Derived ResourceAdequacyMetric class responsible for calculating loss of
    load frequency:

    "[The] expected count of adequacy events per study horizon, with an
    adequacy event defined as a contiguous set of hours with a shortfall

    https://gridops.epri.com/Adequacy/metrics#Loss_of_Load_Events_.28LOLEv.2C_LOLF.29

    Args:
        simulation (ProbabilisticSimulation) : Simulation object to
            characterize.
    """

    def evaluate(self) -> float:
        """Return loss of load frequency

        Returns:
            float: LOLF in units of # of events
        """
        hourly_outages = (
            self.simulation.net_hourly_capacity_matrix < 0
        ).astype(int)
        shifted_hourly_outages = hourly_outages.roll(time=1)
        hourly_outage_transitions = (
            hourly_outages - shifted_hourly_outages
        ) > 0
        return float(
            (hourly_outage_transitions.sum() + hourly_outages[:, -1].sum())
            / hourly_outages.sizes["trial"]
        )



from __future__ import annotations
from datetime import datetime
from logging import getLogger

# external
import numpy as np
import xarray as xr

# package
from assetra.units import NONRESPONSIVE_UNIT_TYPES, RESPONSIVE_UNIT_TYPES
from assetra.system import EnergySystem

LOG = getLogger(__name__)


class ProbabilisticSimulation:
    """Class responsible for creating/storing aggregate probabilistic
    capacity matrices for EnergySystem objects.

    Args:
        start_hour (datetime) : Starting simulation hour,
            for example "2016-01-01 00:00:00"
        end_hour (datetime) : Ending simulation hour (inclusive).
        trial_size (int) : Number of simulation Monte Carlo trials.
    """
    def __init__(
        self, start_hour: datetime, end_hour: datetime, trial_size: int
    ):
        self._start_hour = start_hour
        self._end_hour = end_hour
        self._trial_size = trial_size

        # state variables
        self._energy_system = None
        self._net_hourly_capacity_matrix = None
        self._hourly_capacity_matrix = None

    def copy(self) -> ProbabilisticSimulation:
        """Return a probabilistic simulation object with the same underlying
        parameters but no assigned system

        Returns:
            ProbabilisticSimulation: Simulation with same start hour, end hour,
                and trial size as this object."""
        return ProbabilisticSimulation(
            self._start_hour, self._end_hour, self._trial_size
        )

    def assign_energy_system(self, energy_system: EnergySystem) -> None:
        """Assign an energy system to this probabilistic simulation object.
        Nullifies the stored capacity matrices

        Args:
            energy_system (EnergySystem): Energy system to simulate.

        Raises:
            RuntimeWarning: Invalid type assigned to simulation energy system.
        """
        if not isinstance(energy_system, EnergySystem):
            LOG.warning("Invalid type assigned to simulation energy system.")
            raise RuntimeWarning()
        self._energy_system = energy_system
        self._net_hourly_capacity_matrix = None
        self._hourly_capacity_matrix = None

    @property
    def net_hourly_capacity_matrix(self) -> xr.DataArray:
        """Return a copy of the most recently evaluated net hourly capacity
        matrix

        Returns:
            xr.DataArray: Net hourly capacity matrix with dimensions (trials,
                time) and shape (# of trials, # of hours)

        Raises:
            RuntimeWarning: Net hourly capacity matrix accessed before running simulation
        """
        if self._net_hourly_capacity_matrix is None:
            LOG.error(
                "Net hourly capacity matrix accessed before running simulation"
            )
            raise RuntimeWarning
        return self._net_hourly_capacity_matrix.copy()

    def get_hourly_capacity_matrix_by_type(
        self, unit_type: type
    ) -> xr.DataArray:
        """Return the resultant hourly capacity matrix for a unit_type.

        Hourly capacity matrices for each unit dataset are evaluated by the
        energy unit classes. The probabilistic simulation stores a copy of the
        combined hourly capacity indexed by unit type. It is not possible to
        get the hourly capacity matrix for each unit of a specific type, only
        the aggregate.

        If the hourly capacity matrix does not exist, trigger a simulation run

        Args:
            unit_type (_type_): A valid energy unit type.

        Returns:
            xr.DataArray: Hourly capacity matrix with dimensions (trials, time)
              and shape (# of trials, # of hours)

        Raises:
            RuntimeWarning: Net hourly capacity matrix accessed before running simulation
        """
        if self._hourly_capacity_matrix is None:
            LOG.error(
                "Hourly capacity matrix accessed before running simulation."
            )
            raise RuntimeWarning
        return self._hourly_capacity_matrix.sel(unit_type=unit_type)

    def run(self, net_hourly_capacity_matrix: xr.DataArray = None) -> None:
        """Run the probabilistic simulation. This function evaluates the
        net hourly capacity matrix and hourly capacity matrix for each unit
        type. Optionally, provide a pre-existing net hourly capacity matrix
        to dispatch onto.

        An energy system should already be assigned to the simulation
        object.

        Args:
            net_hourly_capacity_matrix (xr.DataArray, optional): A net hourly
            capacity matrix from a similar simulation with the same start/end
            hours and trial size. If passed, the matrix is modified.
            Defaults to None.

        Raises:
            RuntimeWarning: Energy system not assigned to simulation object.
        """
        # check for energy system
        if not isinstance(self._energy_system, EnergySystem):
            LOG.warning("Energy system not assigned to simulation object.")
            raise RuntimeWarning()

        # setup net hourly capacity matrix
        time_stamps = xr.date_range(
            self._start_hour, self._end_hour, freq="h", inclusive="both"
        )

        # initialize net capacity matrix
        if net_hourly_capacity_matrix is not None:
            # check dimensions
            assert net_hourly_capacity_matrix.sizes["trial"] == self._trial_size
            assert net_hourly_capacity_matrix.sizes["time"] == time_stamps.size
            assert net_hourly_capacity_matrix.time[0] == time_stamps[0]
            # load net hourly capacity
            self._net_hourly_capacity_matrix = net_hourly_capacity_matrix
        else:
            self._net_hourly_capacity_matrix = xr.DataArray(
                data=np.zeros((self._trial_size, len(time_stamps))),
                coords=dict(
                    trial=np.arange(self._trial_size), time=time_stamps
                ),
            )

        # initialize capacity by unit type
        unit_types = NONRESPONSIVE_UNIT_TYPES + RESPONSIVE_UNIT_TYPES
        self._hourly_capacity_matrix = xr.DataArray(
            data=np.zeros(
                (len(unit_types), self._trial_size, len(time_stamps))
            ),
            coords=dict(
                unit_type=unit_types,
                trial=np.arange(self._trial_size),
                time=time_stamps,
            ),
        )

        # iterate through unit datasets
        for unit_type in unit_types:
            if unit_type in self._energy_system.unit_datasets:
                self._hourly_capacity_matrix.loc[
                    unit_type
                ] = unit_type.get_probabilistic_capacity_matrix(
                    self._energy_system.unit_datasets[unit_type],
                    self._net_hourly_capacity_matrix,
                ).values
                self._net_hourly_capacity_matrix += (
                    self._hourly_capacity_matrix.sel(unit_type=unit_type).values
                )



from logging import getLogger
from abc import ABC, abstractmethod

from assetra.units import RESPONSIVE_UNIT_TYPES, NONRESPONSIVE_UNIT_TYPES
from assetra.system import EnergySystem
from assetra.simulation import ProbabilisticSimulation
from assetra.metrics import ResourceAdequacyMetric

import xarray as xr

LOG = getLogger(__name__)


class ResourceContributionMetric(ABC):
    """Class responsible for quantifying resource contributions
    to an energy system

    Args:
        energy_system (EnergySystem): Base system to which resources are added
        simulation (ProbabilisticSimulation): Instantiated simulation object
            whose parameters (i.e. time range and trial size are used
            throughout the ELCC calculation)
        resource_adequacy_metric: Class derived
            from ResourceAdequacyMetric to use throughout the ELCC simulation
            (e.g. ExpectedUnservedEnergy, LossOfLoadHours)
    """

    def __init__(
        self,
        energy_system: EnergySystem,
        simulation: ProbabilisticSimulation,
        resource_adequacy_metric: type[ResourceAdequacyMetric],
    ):
        self._original_energy_system = energy_system
        self._simulation = simulation
        self._resource_adequacy_metric = resource_adequacy_metric

    @abstractmethod
    def evaluate(self, addition: EnergySystem) -> float:
        """Return resource contribution of addition to the energy system

        Args:
            addition (EnergySystem): Energy system to add (possibly a single
            unit)

        Returns:
            float: Quantified resource contribution of addition to the energy
                system.
        """


class EffectiveLoadCarryingCapability(ResourceContributionMetric):
    """Class responsible for quantifying resource contribution to an energy
    system using the effective load-carrying capability metric:

    "[The] amount by which the system's load can increase when the resource is
    added to the system while maintaining the same system adequacy."

    https://gridops.epri.com/Adequacy/metrics#Effective_Load_Carrying_Capability_.28ELCC.29

    Args:
        energy_system (EnergySystem) : Base system to which resources are added
        simulation (ProbabilisticSimulation) : Instantiated simulation object
            whose parameters (i.e. time range and trial size are used
            throughout the ELCC calculation)
        resource_adequacy_metric (type[ResourceAdequacyMetric') : Class derived
            from ResourceAdequacyMetric to use throughout the ELCC simulation
            (e.g. ExpectedUnservedEnergy, LossOfLoadHours)
    """

    def __init__(
        self,
        energy_system: EnergySystem,
        simulation: ProbabilisticSimulation,
        resource_adequacy_metric: type[ResourceAdequacyMetric],
    ):
        """The ELCC computes resource adequacy many times iteratively. In many
        instances, it is neither necessary nor desirable to re-run the
        probabilistic simulation for a set of units more than once. For
        so-called non-responsive units (static and stochastic units) which do
        not respond to system conditions, a single probabilistic simulation
        is sufficient. For responsive units, including storage, changing system
        conditions necessitate re-evaluation of probabilistic capacity
        contributions.

        To account for this classification in the iterative resource adequacy
        calculations, the ELCC decomposes energy systems into their responsive
        and non-responsive components. In each iteration, combinations of the
        component energy systems' probabilistic simulations are "chained"
        together depending on which need to be recomputed.

        For example, to establish the baseline resource adequacy (as in the
        init function), both the original responsive system and the
        original non-responsive systems are evaluated. Programmatically, the
        non-responsive system is evaluated first. Then the resulting
        net hourly capacity matrix is fed into the responsive system.

        When a new system is added, it is again decomposed into responsive and
        non-responsive subsystems. The original non-responsive system does not
        need to be re-evaluated, so its already-simulated probabilistic net
        hourly capacity is fed into the new non-responsive system which is fed
        into the new and original responsive systems respectively.

        In this way, non-responsive units are only evaluated once while
        responsive units are able to respond to changes in system conditions.
        """
        ResourceContributionMetric.__init__(
            self, energy_system, simulation, resource_adequacy_metric
        )

        # decompose system into responsive and non-responsive components
        # non-responsive simulation
        self._original_system_non_responsive = (
            self._original_energy_system.get_system_by_type(
                NONRESPONSIVE_UNIT_TYPES
            )
        )
        self._original_non_responsive_simulation = self._simulation.copy()
        self._original_non_responsive_simulation.assign_energy_system(
            self._original_system_non_responsive
        )

        # responsive simulation
        self._original_system_responsive = (
            self._original_energy_system.get_system_by_type(
                RESPONSIVE_UNIT_TYPES
            )
        )
        self._original_responsive_simulation = self._simulation.copy()
        self._original_responsive_simulation.assign_energy_system(
            self._original_system_responsive
        )

        # run chained simulation
        self._original_non_responsive_simulation.run()
        self._original_responsive_simulation.run(
            self._original_non_responsive_simulation.net_hourly_capacity_matrix
        )
        self._resource_adequacy_model = self._resource_adequacy_metric(
            self._original_responsive_simulation
        )
        self._original_resource_adequacy = (
            self._resource_adequacy_model.evaluate()
        )
        LOG.info(
            "Original resource adequacy: "
            + str(self._original_resource_adequacy)
        )

        if self._original_resource_adequacy == 0:
            LOG.error("Invalid ELCC calculation for system with no risk")
            raise RuntimeWarning()

        # save intermediate steps
        self._original_net_capacity_matrix = (
            self._original_responsive_simulation.net_hourly_capacity_matrix
        )
        self._intermediate_net_capacity_matrices = []

    @property
    def original_net_capacity_matrix(self) -> xr.DataArray:
        """Return the net hourly capacity matrix of the base system"""
        # TODO test
        return self._original_net_capacity_matrix.copy()

    @property
    def intermediate_net_capacity_matrices(
        self,
    ) -> tuple[tuple[float, xr.DataArray]]:
        """Return intermediate net capacity matrices for each step of the ELCC
        calculation. Each element is a tuple composed of the amount of added
        constant load and the net hourly capacity matrix corresponding to that
        step.
        """
        # TODO test
        return tuple(m for m in self._intermediate_net_capacity_matrices)

    def evaluate(
        self,
        addition: EnergySystem,
        additional_demand_resolution_pct: float = 0.01,
    ) -> float:
        """Return the ELCC of an addition to the energy system.

        Args:
            addition (EnergySystem): Energy system to add (possibly a
                single unit).
            additional_demand_resolution_pct (float, optional): Resolution of
                added demand to find as percent of added nameplate capacity.
                Defaults to 0.01 (1%). E.g. for a 100 MW addition, the ELCC
                will be found within 1 MW.

        Returns:
            float: Amount of added constant load in units of power.
        """

        # reset intermediate net capacities
        self._intermediate_net_capacity_matrices = []

        # decompose system into responsive and non-responsive components
        # non-responsive simulation
        additional_system_non_responsive = addition.get_system_by_type(
            NONRESPONSIVE_UNIT_TYPES
        )
        additional_non_responsive_simulation = self._simulation.copy()
        additional_non_responsive_simulation.assign_energy_system(
            additional_system_non_responsive
        )

        # responsive simulation
        additional_system_responsive = addition.get_system_by_type(
            RESPONSIVE_UNIT_TYPES
        )
        additional_responsive_simulation = self._simulation.copy()
        additional_responsive_simulation.assign_energy_system(
            additional_system_responsive
        )

        # run non-responsive_simulation
        additional_non_responsive_simulation.run()

        # get non-responsive net hourly capacity
        non_responsive_net_hourly_capacity_matrix = (
            self._original_non_responsive_simulation.net_hourly_capacity_matrix
            + additional_non_responsive_simulation.net_hourly_capacity_matrix
        )

        # add load
        additional_demand_upper_bound = addition.system_capacity
        additional_demand_lower_bound = 0
        additional_demand_resolution = (
            additional_demand_upper_bound - additional_demand_lower_bound
        )
        additional_demand = (
            additional_demand_lower_bound + additional_demand_resolution / 2
        )

        # run chained responsive simulation
        additional_responsive_simulation.run(
            non_responsive_net_hourly_capacity_matrix - additional_demand
        )
        self._original_responsive_simulation.run(
            additional_responsive_simulation.net_hourly_capacity_matrix
        )
        self._intermediate_net_capacity_matrices.append(
            (
                additional_demand,
                self._original_responsive_simulation.net_hourly_capacity_matrix,
            )
        )

        # update resource adequacy
        new_resource_adequacy = self._resource_adequacy_model.evaluate()

        # printout
        LOG.info("Additional demand: " + str(round(additional_demand)))
        LOG.info("Resource adequacy: " + str(new_resource_adequacy))

        # iterate until convergence
        iteration = 0

        while (
            additional_demand_resolution / addition.system_capacity
        ) > additional_demand_resolution_pct:
            # iterate until original resource adequacy level is met
            if new_resource_adequacy > self._original_resource_adequacy:
                # if over-reliable, add load
                additional_demand_upper_bound = additional_demand
            else:
                # if under-reliable, remove load
                additional_demand_lower_bound = additional_demand
            # add demand
            additional_demand_resolution = (
                additional_demand_upper_bound - additional_demand_lower_bound
            )
            additional_demand = (
                additional_demand_lower_bound + additional_demand_resolution / 2
            )

            # run chained responsive simulation
            additional_responsive_simulation.run(
                non_responsive_net_hourly_capacity_matrix - additional_demand
            )
            self._original_responsive_simulation.run(
                additional_responsive_simulation.net_hourly_capacity_matrix
            )
            self._intermediate_net_capacity_matrices.append(
                (
                    additional_demand,
                    self._original_responsive_simulation.net_hourly_capacity_matrix,
                )
            )

            # update resource adequacy
            new_resource_adequacy = self._resource_adequacy_model.evaluate()

            LOG.info("Additional demand: " + str(additional_demand))
            LOG.info("Resource adequacy: " + str(new_resource_adequacy))

            # update iteration count
            iteration += 1

        return float(additional_demand)
